package kirjasto;

import fi.jyu.mit.ohj2.Mjonot;
import kanta.NumeroTarkistus;

import java.io.*;

/**
 * Arvostelu, joka huolelhtii ID:sta
 * @author Erno Pajala
 * @version 08.03.2025
 */
public class Arvostelu implements Cloneable {
    // Muuttujat
    private int arvosteluID;
    private int elokuvaID;
    private String arvosana;
    private String kommentti = "";


    /**
     * Muodostaja
     */
    public Arvostelu() {
        // Ei tarvita viela
    }


    /**
     * Alustetaan tietyn elokuvan arvostelu
     * @param elokuvaID elokuvaID
     */
    public Arvostelu(int elokuvaID) {
        this.elokuvaID = elokuvaID;
    }


    // Getterit ja setterit
    /**
     * @return Elokuvan ID
     */
    public int getElokuvaID() {
        return elokuvaID;
    }

    /**
     * @return Elokuvan arvosana
     */
    public String getArvosana() {
        return arvosana;
    }


    /**
     * @return ArvostelunID
     */
    public int getArvosteluID() {
        return arvosteluID;
    }


    /**
     * @return Elokuvan komentti
     * @example
     * <pre name="test">
     *   int movieID = 1;
     *   Arvostelu arv1 = new Arvostelu(movieID);
     *   arv1.taytaArvostelu(movieID);
     *   arv1.getKommentti() =R= "Hyva leffa.*";
     * </pre>
     */
    public String getKommentti() {
        return kommentti;
    }


    /**
     * Asetetaan uusi arvosana
     * @param arvosana arvosana
     * @return jos ok niin null, muuten virhe
     */
    public String setArvosana(String arvosana) {

        String virhe = NumeroTarkistus.tarkista(arvosana, true);
        // Jos saadaan virhe palautetaan se, eika vaiheta arvosanaa
        if (virhe != null) {
            return virhe;
        }

        this.arvosana = arvosana;
        return null;
    }


    /**
     * Asetetaan uusi kommentti
     * @param kommentti kommentti
     */
    public void setKommentti(String kommentti) {
        this.kommentti = kommentti;
    }


    /**
     * Asettaa arvosteluIDn
     * @param aid asetettava arvosteluID
     */
    private void setArvosteluID(int aid) {
        arvosteluID = aid;
    }


    /**
     * Antaa arvostelulle seuraavan rekisterinumeron.
     * @return arvostusten uusi ID
     * @example
     * <pre name="test">
     *   Arvostelu arv1 = new Arvostelu();
     *   arv1.getArvosteluID() === 0;
     *   arv1.rekisteroi(1);
     *   Arvostelu arv2 = new Arvostelu();
     *   arv2.rekisteroi(2);
     *   int n1 = arv1.getArvosteluID();
     *   int n2 = arv2.getArvosteluID();
     *   n1 === n2-1;
     * </pre>
     */

    public int rekisteroi(int eID) {
        arvosteluID = eID;
        elokuvaID = arvosteluID;
        return arvosteluID;
    }



    /**
     * Apumetodi, jolla saadaan täytettyä testiarvot arvostelulle
     * @param nro ElokuvaID
     */
    public void taytaArvostelu(int nro) {
        elokuvaID = nro;
        arvosana = "10";
        kommentti = "Hyva leffa";
    }


    /**
     * Tulostetaab elokuvan tiedot
     * @param out tietovirta johon tulostetaan
     */
    public void tulosta(PrintStream out) {
        out.println("Arvostelu ID: " + arvosteluID);
        out.println("Elokuvan ID: " + arvosteluID);
        out.println("Kommentti: " + kommentti);
        out.println("Arvosana: " + arvosana);
    }


    /**
     * Tulostetaan arvostelun tiedot
     * @param os tietovirta johon tulostetaan
     */
    public void tulosta(OutputStream os) {
        tulosta(new PrintStream(os));
    }


    /**
     * Selvitaa arvostelun tiedot | erotellusta merkkijonosta ja lisaa ne arvostelulle
     * Pitaa huolen etta seuraava ID on suurempi kuin tuleva ID
     * @param rivi josta elokuvan tiedot otetaan
     * @example
     * <pre name="test">
     *   Arvostelu arvostelu = new Arvostelu();
     *   arvostelu.parse(" 2 | 2 | 4");
     *   arvostelu.getArvosteluID() === 2;
     *   arvostelu.toString().startsWith("2 | 2 | 4 |") === true;
     *   int n = arvostelu.getArvosteluID();
     *   arvostelu.parse(""+(n+20)); // Otetaan merkkijonosta vain arvosteluID
     *   arvostelu.getArvosteluID() === n+20;
     * </pre>
     */
    public void parse(String rivi){
        var sb = new StringBuilder(rivi);
        // Koitetaan erottaa, jos ei pystyta kaytetaan alkuperaista arvoa
        setArvosteluID(Mjonot.erota(sb, '|', getArvosteluID())); // Varmistetaan, etta suraava ID on aina suurempi
        elokuvaID = Mjonot.erota(sb, '|', elokuvaID);
        arvosana = Mjonot.erota(sb, '|', arvosana);
        kommentti = Mjonot.erota(sb, '|', kommentti);
    }


    /**
     * Palauttaa arvostelun teidot merkkijonona jonka voi tallentaa tiedostoon
     * @return arvostelun tiedot
     * @example <pre name="test">
     * Arvostelu arv = new Arvostelu();
     * arv.parse("1 | 1 | 10");
     * arv.toString().startsWith(1 | 1 | 10 |") === true;
     * </pre>
     */
    @Override
    public String toString() {
        return getArvosteluID() + " | " +
                getElokuvaID() + " | " +
                getArvosana() + " | " +
                getKommentti();
    }


    /**
     * Tehdaan klooni arvostelusta
     * @return Kloonattu arvostelu
     * @throws CloneNotSupportedException
     * @example <pre name="test">
     * #THROWS CloneNotSupportedException
     *   Arvostelu arvostelu = new Arvostelu();
     *   arvostelu.parse("1 | 1 | 10");
     *   Arvostelu kopio = arvostelu.clone();
     *   kopio.toString() === arvostelu.toString();
     *   arvostelu.parse("2 | 2 | 8");
     *   kopio.toString().equals(arvostelu.toString()) === false;
     * </pre>
     */
    @Override
    public Arvostelu clone() throws CloneNotSupportedException {
        Arvostelu uusi;
        uusi = (Arvostelu) super.clone();
        return uusi;
    }


    /**
     * Testiohjelma arvosteluille
     * @param args Ei kaytossa
     */
    public static void main(String[] args) {
        Arvostelu arv = new Arvostelu();
        arv.taytaArvostelu(1);
        arv.tulosta(System.out);
    }
}
