package kirjasto;

import java.io.*;
import java.util.*;
/**
 * Arvostelut, lisaa uuden arvostelun
 * @author Erno Pajala
 * @version 08.03.2025
 **/
public class Arvostelut implements Iterable<Arvostelu> {

    private boolean muutettu = false;
    /** Taulukko arvosteluista */
    private final Collection<Arvostelu> alkiot = new ArrayList<>();
    // Käytetaan yksinkertaista (collection) omisaisuutta arraylist rakenteesta


    /**
     * Arvostelun alustaminen
     */
    public Arvostelut() {
        // toistaiseksi ei tarvitse tehdä mitään
    }


    /**
     * Lisää uuden arvostelun tietorakenteeseen.  Ottaa arvostelun omistukseensa.
     * @param arv lisättävä arvostelu. Huom tietorakenne muuttuu omistajaksi
     */
    public void lisaa(Arvostelu arv) {
        alkiot.add(arv);
        muutettu = true;
    }


    /**
     * Korvaa arvostelun.
     * Etsitaan samalla ID:lla oleva arvostelu, jos ei löydy, niin
     * lisataan uutena arvosteluna.
     * @param korvattavaArvostelu lisattavan elokuvan viite. Tietorakenne muuttuu omistajaksi
     * @throws SailoException jos tietorakenne on taynna
     * @example <pre name="test">
     * #THROWS SailoException,CloneNotSupportedException
     * #PACKAGEIMPORT
     * Arvostelut arvostelut = new Arvostelut();
     * Arvostelu a1 = new Arvostelu(), a2 = new Arvostelu();
     * a1.rekisteroi(1); a2.rekisteroi(2);
     * arvostelut.korvaaTaiLisaa(a1);
     * arvostelut.korvaaTaiLisaa(a2);
     * Arvostelu a3 = a1.clone();
     * a3.setKommentti("Hyva leffa");
     * Iterator<Arvostelu> it = arvostelut.iterator();
     * it.next() == a1 === true;
     * it = arvostelut.iterator();
     * Arvostelu a0 = it.next();
     * a0 == a3 === false;
     * a0 == a1 === true;
     * </pre>
     */
    public void korvaaTaiLisaa(Arvostelu korvattavaArvostelu) throws SailoException {
        int eID = korvattavaArvostelu.getElokuvaID();

        // Käydään läpi kaikki arvostelut ja etsitään sama elokuvaID
        Iterator<Arvostelu> it = alkiot.iterator();
        while (it.hasNext()) {
            Arvostelu arv = it.next();
            if (arv.getElokuvaID() == eID) {
                // Poistetaan vanha ja lisätään uusi
                it.remove();
                alkiot.add(korvattavaArvostelu);
                muutettu = true;
                return;
            }
        }

        // Jos ei loydy lisataan uutena arvosteluna
        lisaa(korvattavaArvostelu);
    }


    /**
     * Poistetaan arvostelu elokuvaIDn avulla
     * @param eID elokuvaID
     */
    public void poista(int eID) {
        Iterator<Arvostelu> it = alkiot.iterator();
        while (it.hasNext()) {
            Arvostelu arv = it.next();
            if (arv.getElokuvaID() == eID) {
                // Poistetaan vanha ja lisätään uusi
                it.remove();
                muutettu = true;
                return;
            }
        }
    }


    /**
     * Lukee arvostelut tiedostosta
     * @param hakemisto tiedoston hakemisto
     * @throws SailoException jos lukeminen epäonnistuu
     * @example
     * <pre name="test">
     * #THROWS SailoException
     * #import java.io.*;
     * // Testaa tallennus ja lukeminen
     * String testiHakemisto = "testidata";
     * Arvostelut arvostelut = new Arvostelut();
     * Arvostelu a1 = new Arvostelu(), a2 = new Arvostelu();
     * a1.rekisteroi(1); a1.taytaArvostelu(1);
     * a2.rekisteroi(2); a2.taytaArvostelu(2);
     * arvostelut.lisaa(a1);
     * arvostelut.lisaa(a2);
     * // Testaa kun tiedostoa ei voi lukea
     * arvostelut.lueTiedostosta(testiHakemisto); #THROWS SailoException
     * arvostelut.tallenna(testiHakemisto);
     * // Lue tiedot uusiin arvosteluihin
     * Arvostelut uudet = new Arvostelut();
     * uudet.lueTiedostosta(testiHakemisto);
     * Iterator<Arvostelu> i = uudet.iterator();
     * i.next().toString() === a1.toString();
     * i.next().toString() === a2.toString();
     * i.hasNext() === false;
     * // Siivoa testitiedosto
     * new File(testiHakemisto + "/arvostelut.dat").delete();
     * new File(testiHakemisto).delete();
     * </pre>
     */
    public void lueTiedostosta(String hakemisto) throws SailoException {
        String tiedostonNimi = hakemisto + "/arvostelut.dat";
        File ftiedosto = new File(tiedostonNimi);

        if (!ftiedosto.exists()) {
            throw new SailoException("Tiedostoa ei löydy: " + ftiedosto.getAbsolutePath());
        }

        try (Scanner scanner = new Scanner(new FileInputStream(ftiedosto))) {
            while ( scanner.hasNext() ) { // Jos seuraava olemassa jatka lukemista
                String s = scanner.nextLine();
                if (s ==  null || s.isEmpty()) continue; // Katsotaan lopetetaanko silmukka
                Arvostelu arvostelu = new Arvostelu();
                arvostelu.parse(s); // Otetaan rivilta tiedot ja laitetaan ne arvosteluun
                lisaa(arvostelu);
            }
            muutettu = false;
        } catch ( FileNotFoundException e ) {
            throw new SailoException("Ei saa luettua tiedostoa " + tiedostonNimi);
        }
    }


    /**
     * Tallentaa arvostelut tiedostoon.
     * @param hakemisto Mihin hakemistoon tallennetaan
     * @throws SailoException jos talletus epäonnistuu
     */
    public void tallenna(String hakemisto) throws SailoException {
        if ( !muutettu ) return; // Ei tallenneta jos ei olla muutettu

        File kansio = new File(hakemisto);
        if (!kansio.exists()) {
            if (!kansio.mkdirs()) {
                throw new SailoException("Hakemiston luominen epäonnistui: " + kansio.getAbsolutePath());
            }
        }

        File ftiedosto = new File(hakemisto + "/arvostelut.dat");
        try (PrintStream ps = new PrintStream(new FileOutputStream(ftiedosto)) ) {

            // Kaydaan lapi kaikkien elokuvien arvostelut ja laitetaan ne rivi kerrallaan tiedostoon
            for (Arvostelu arv : alkiot) {
                ps.println(arv.toString());
            }

        } catch (FileNotFoundException e) {
            throw new SailoException("Tiedosto " + ftiedosto.getAbsolutePath());
        }
        
        muutettu = false;
    }


    /**
     * Iteraattori kaikkien arvostelujen läpikäymiseen
     * @return arvosteluitelaattori
     * @example
     * <pre name="test">
     * #PACKAGEIMPORT
     * #import java.util.*;
     *  Arvostelut arvosteet = new Arvostelut();
     *  Arvostelu arv21 = new Arvostelu(2); arvosteet.lisaa(arv21);
     *  Arvostelu arv11 = new Arvostelu(1); arvosteet.lisaa(arv11);
     *  Arvostelu arv22 = new Arvostelu(2); arvosteet.lisaa(arv22);
     *  Arvostelu arv12 = new Arvostelu(1); arvosteet.lisaa(arv12);
     *  Arvostelu arv23 = new Arvostelu(2); arvosteet.lisaa(arv23);
     *  Iterator<Arvostelu> i2=arvosteet.iterator();
     *  i2.next() === arv21;
     *  i2.next() === arv11;
     *  i2.next() === arv22;
     *  i2.next() === arv12;
     *  i2.next() === arv23;
     *  i2.next() === arv12;  #THROWS NoSuchElementException
     *  int n = 0;
     *  int eID[] = {2,1,2,1,2};
     *  for ( Arvostelu arv : arvosteet ) {
     *    arv.getElokuvaID() === eID[n]; n++;
     *  }
     *
     *  n === 5;
     *
     */
    @Override
    public Iterator<Arvostelu> iterator() {
        return alkiot.iterator();
    }


    /**
     * Haetaan kaikki elokuvan arvostelut
     * @param elokuvaID elokuvan ID jolle arvosteluja haetaan
     * @return Arvostelu joka löydettiin, muuten null
     *
     * <pre name="test">
     * #import java.util.*;
     *  Arvostelut arvosteet = new Arvostelut();
     *  Arvostelu arv21 = new Arvostelu(1); arvosteet.lisaa(arv21);
     *  Arvostelu arv11 = new Arvostelu(2); arvosteet.lisaa(arv11);
     *  Arvostelu arv22 = new Arvostelu(4); arvosteet.lisaa(arv22);
     *  arvosteet.annaArvostelu(1) === arv21;
     *  arvosteet.annaArvostelu(2) === arv11;
     *  arvosteet.annaArvostelu(3) === null;
     *  arvosteet.annaArvostelu(4) === arv22;
     * </pre>
     */
    public Arvostelu annaArvostelu(int elokuvaID) {
        for (Arvostelu arv : alkiot) // Kaydaan kaikki arvostelut läpi ja katsotaan löytyykö annettu elokuvaID
            if (arv.getElokuvaID() == elokuvaID) { // Palautetaan arvostelu, jos silla on sama elokuvaID
                return arv;
            }
        return null;
    }


    /**
     * Testiohjelma arvosteluille
     * @param args ei käytössä
     */
    public static void main(String[] args) {
        Arvostelut arvosteet = new Arvostelut();

        try {
            arvosteet.lueTiedostosta("omakirjasto");

            Arvostelu arv1 = new Arvostelu();
            arv1.rekisteroi(1);
            arv1.taytaArvostelu(3);
            arvosteet.lisaa(arv1);

            arvosteet.tallenna("omakirjasto");
        } catch (SailoException e) {
            throw new RuntimeException(e);
        }

        System.out.println("============= Arvostelujen testi =================");
        for (Arvostelu arv : arvosteet) {
            arv.tulosta(System.out);
        }
    }
}
