package kirjasto;

import java.io.*;
import java.util.*;

/**
 * Elokuvat luokka, pystyy kutusmaan ja muokkaamaan yksittaisia elokuvia
 *
 * @author Erno Pajala
 * @version 26.02.2025
 */
public class Elokuvat implements Iterable<Elokuva> {

    private static final int MAX_ELOKUVIA = 10; // Kuinka monta alkiota ensin syntyy
    private int lkm = 0; // Kuinka monta taytettya alkiota ensin
    private Elokuva[] alkiot = new Elokuva[MAX_ELOKUVIA];
    private boolean muutettu = false; // Katsotaan onko tehty muutoksia


    /**
     * Luodaan alustava taulukko
     */
    public Elokuvat() {
        //
    }


    /**
     * Palauttaa kirjaston elokuvien lukumaaran
     * @return elokuvien lukumaara
     */
    public int getLkm() {
        return lkm;
    }


    /**
     * Lisää uuden elokuvan tietorakenteeseen.  Ottaa elokuvan omistukseensa.
     * @param elokuva lisätäävän elokuvaan viite.  Huom tietorakenne muuttuu omistajaksi
     * @throws SailoException jos tietorakenne on jo täynnä
     * @example
     * <pre name="test">
     * #THROWS SailoException
     * Elokuvat elokuvat = new Elokuvat();
     * Elokuva Terminator = new Elokuva(), Terminator2 = new Elokuva();
     * elokuvat.getLkm() === 0;
     * elokuvat.lisaa(Terminator); elokuvat.getLkm() === 1;
     * elokuvat.lisaa(Terminator2); elokuvat.getLkm() === 2;
     * elokuvat.lisaa(Terminator); elokuvat.getLkm() === 3;
     * elokuvat.anna(0) === Terminator;
     * elokuvat.anna(1) === Terminator2;
     * elokuvat.anna(2) === Terminator;
     * elokuvat.anna(1) == Terminator === false;
     * elokuvat.anna(1) == Terminator2 === true;
     * elokuvat.anna(3) === Terminator; #THROWS IndexOutOfBoundsException
     * elokuvat.lisaa(Terminator); elokuvat.getLkm() === 4;
     * elokuvat.lisaa(Terminator); elokuvat.getLkm() === 5;
     * </pre>
     */
    public void lisaa(Elokuva elokuva) throws SailoException {
        // kasvatetaan listaa kun tietorakenne taynna
        if (lkm >= alkiot.length) {
            alkiot = Arrays.copyOf(alkiot, lkm + 10);
        }
        alkiot[lkm] = elokuva;
        lkm++;
        muutettu = true;
    }


    /**
     * Korvaa elokuvan tietorakenne. Ottaa elokuvan omistukseensa.
     * Etsitaan samalla ID:lla oleva elokuva, jos ei löydy, niin
     * lisataan uutena elokuvana.
     * @param elokuva lisattavan elokuvan viite. Tietorakenne muuttuu omistajaksi
     * @throws SailoException jos tietorakenne on taynna
     * @example <pre name="test">
     * #THROWS SailoException,CloneNotSupportedException
     * #PACKAGEIMPORT
     * Elokuvat elokuvat = new Elokuvat();
     * Elokuva e1 = new Elokuva(), e2 = new Elokuva();
     * e1.rekisteroi(); e2.rekisteroi();
     * elokuvat.getLkm() === 0;
     * elokuvat.korvaaTaiLisaa(e1); elokuvat.getLkm() === 1;
     * elokuvat.korvaaTaiLisaa(e2); elokuvat.getLkm() === 2;
     * Elokuva e3 = e1.clone();
     * e3.setName("Pulp Fiction");
     * Iterator<Elokuva> it = elokuvat.iterator();
     * it.next() == e1 === true;
     * elokuvat.korvaaTaiLisaa(e3); elokuvat.getLkm() === 2;
     * it = elokuvat.iterator();
     * Elokuva e0 = it.next();
     * e0 === e3;
     * e0 == e3 === true;
     * e0 == e1 === false;
     * </pre>
     */
    public void korvaaTaiLisaa(Elokuva elokuva) throws SailoException {
        int id = elokuva.getElokuvaID();
        // kaydaan lapi kaikki elokuvat ja vaihetaan muokattava elokuva uuteen elokuvaID:n avulla
        for (int i = 0; i < lkm; i++) {
            if (alkiot[i].getElokuvaID() == id) {
                alkiot[i] = elokuva;
                muutettu = true;
                return;
            }
        }

        // Jos ei loydy lisataan uutena elokuvana
        lisaa(elokuva);
    }


    /**
     * Palauttaa viitteen i:teen elokuvaan.
     * @param i monennenko elokuvan viite halutaan
     * @return viite elokuvaan, jonka indeksi on i
     * @throws IndexOutOfBoundsException jos i ei ole sallitulla alueella
     */
    public Elokuva anna(int i) throws IndexOutOfBoundsException {
        if (i < 0 || lkm <= i)
            throw new IndexOutOfBoundsException("Laiton indeksi: " + i);
        return alkiot[i];
    }


    /**
     * Etsii elokuva id:n perusteella
     * @param id elokuvanID, jonka mukaan etsitään
     * @return elokuva jolla etsittävä id tai null
     * <pre name="test">
     * #THROWS SailoException
     * Elokuvat elokuvat = new Elokuvat();
     * Elokuva e1 = new Elokuva(), e2 = new Elokuva(), e3 = new Elokuva();
     * e1.rekisteroi(); e2.rekisteroi(); e3.rekisteroi();
     * int id1 = e1.getElokuvaID();
     * elokuvat.lisaa(e1); elokuvat.lisaa(e2); elokuvat.lisaa(e3);
     * elokuvat.annaId(id1  ) == e1 === true;
     * elokuvat.annaId(id1+1) == e2 === true;
     * elokuvat.annaId(id1+2) == e3 === true;
     * </pre>
     */
    public Elokuva annaId(int id) {
        for (Elokuva elokuva : this) {
            if (id == elokuva.getElokuvaID()) return elokuva;
        }
        return null;
    }


    /**
     * Etsii elokuvan id:n perusteella
     * @param id elokuvanID, jonka mukaan etsitään
     * @return löytyneen jäsenen indeksi tai -1 jos ei löydy
     * <pre name="test">
     * #THROWS SailoException
     * Elokuvat elokuvat = new Elokuvat();
     * Elokuva e1 = new Elokuva(), e2 = new Elokuva(), e3 = new Elokuva();
     * e1.rekisteroi(); e2.rekisteroi(); e3.rekisteroi();
     * int id1 = e1.getElokuvaID();
     * elokuvat.lisaa(e1); elokuvat.lisaa(e2); elokuvat.lisaa(e3);
     * elokuvat.etsiId(id1+1) === 1;
     * elokuvat.etsiId(id1+2) === 2;
     * </pre>
     */
    public int etsiId(int id) {
        for (int i = 0; i < lkm; i++)
            if (id == alkiot[i].getElokuvaID()) return i;
        return -1;
    }


    /**
     * Lukee elokuvien tiedostosta
     * @param hakemisto tiedoston hakemisto
     * @throws SailoException jos lukeminen epäonnistuu
     * @example
     * <pre name="test">
     * #THROWS SailoException
     * #import java.io.*;
     * // Testaa tallennus ja lukeminen
     * String testiHakemisto = "testidata";
     * Elokuvat elokuvat = new Elokuvat();
     * Elokuva e1 = new Elokuva(), e2 = new Elokuva();
     * e1.rekisteroi(); e1.taytaElokuva();
     * e2.rekisteroi(); e2.taytaElokuva();
     * elokuvat.lisaa(e1);
     * elokuvat.lisaa(e2);
     * // Testaa kun tiedostoa ei voi lukea
     * elokuvat.lueTiedostosta(testiHakemisto); #THROWS SailoException
     * elokuvat.tallenna(testiHakemisto);
     * // Lue tiedot uusiin elokuviin
     * Elokuvat uudet = new Elokuvat();
     * uudet.lueTiedostosta(testiHakemisto);
     * uudet.getLkm() === 2;
     * uudet.anna(0).getOhjaaja() === "James Cameron";
     * Iterator<Elokuva> i = uudet.iterator();
     * i.next().toString() === e1.toString();
     * i.next().toString() === e2.toString();
     * i.hasNext() === false;
     * // Siivoa testitiedosto
     * new File(testiHakemisto + "/elokuvat.dat").delete();
     * new File(testiHakemisto).delete();
     * </pre>
     */
    public void lueTiedostosta(String hakemisto) throws SailoException {
        String tiedostonNimi = hakemisto + "/elokuvat.dat";
        File ftiedosto = new File(tiedostonNimi);

        if (!ftiedosto.exists()) {
            throw new SailoException("Tiedostoa ei löydy: " + ftiedosto.getAbsolutePath());
        }

        try (Scanner scanner = new Scanner(new FileInputStream(ftiedosto))) {
            while ( scanner.hasNext() ) { // Jos seuraava olemassa jatka lukemista
                String s = scanner.nextLine();
                if (s ==  null || s.isEmpty()) continue; // Katsotaan lopetetaanko silmukka
                Elokuva elokuva = new Elokuva();
                elokuva.parse(s); // Otetaan rivilta tiedot ja laitetaan ne elokuvaan
                lisaa(elokuva);
            }
            muutettu = false;

        } catch ( FileNotFoundException e ) {
            throw new SailoException("Ei saa luettua tiedostoa " + tiedostonNimi);
        }
    }


    /**
     * Tallentaa elokuvat tiedostoon.
     * @param hakemisto Mihin hakemistoon tallennetaan
     * @throws SailoException jos talletus epäonnistuu
     */
    public void tallenna(String hakemisto) throws SailoException {
        if ( !muutettu) return; // Jos ei ole tehty muutoksia ei tallenneta

        File kansio = new File(hakemisto);
        if (!kansio.exists()) {
            if (!kansio.mkdirs()) {
                throw new SailoException("Hakemiston luominen epäonnistui: " + kansio.getAbsolutePath());
            }
        }

        File ftiedosto = new File(hakemisto + "/elokuvat.dat");
        try (PrintStream ps = new PrintStream(new FileOutputStream(ftiedosto)) ) {
            // kaydaan lapi kaikki elokuvat ja kirjoitetaan ne tiedostoon
            for (int i = 0; i < this.getLkm(); i++) {
                Elokuva elokuva = this.anna(i);
                ps.println(elokuva.toString());
            }
        } catch (FileNotFoundException e) {
            throw new SailoException("Tiedosto " + ftiedosto.getAbsolutePath());
        }

        muutettu = false;
    }


    /**
     * Poistaa elokuva jolla on valittu ID
     * @param id poistettavan elokuvan ID
     * @return 1 jos poistettiin, 0 jos ei löydy
     * @example
     * <pre name="test">
     * #THROWS SailoException
     * Elokuvat elokuvat = new Elokuvat();
     * Elokuva e1 = new Elokuva(), e2 = new Elokuva(), e3 = new Elokuva();
     * e1.rekisteroi(); e2.rekisteroi(); e3.rekisteroi();
     * int id1 = e1.getElokuvaID();
     * elokuvat.lisaa(e1); elokuvat.lisaa(e2); elokuvat.lisaa(e3);
     * elokuvat.poista(id1+1) === 1;
     * elokuvat.annaId(id1+1) === null; elokuvat.getLkm() === 2;
     * elokuvat.poista(id1) === 1; elokuvat.getLkm() === 1;
     * elokuvat.poista(id1+3) === 0; elokuvat.getLkm() === 1;
     * </pre>
     */
    public int poista(int id) {
        int ind = etsiId(id);
        if (ind < 0) return 0;
        lkm--;
        for (int i = ind; i < lkm; i++)
            alkiot[i] = alkiot[i + 1];
        alkiot[lkm] = null;
        muutettu = true;
        return 1;
    }

    /**
     * Luokka elokuvien iteroimiseksi
     * @example
     * <pre name="test">
     * #THROWS SailoException
     * #PACKAGEIMPORT
     * #import java.util.*;
     * Elokuvat elokuvat = new Elokuvat();
     * Elokuva terminator1 = new Elokuva(), terminator2 = new Elokuva();
     * terminator1.rekisteroi(); terminator2.rekisteroi();
     * elokuvat.lisaa(terminator1);
     * elokuvat.lisaa(terminator2);
     * elokuvat.lisaa(terminator1);
     * StringBuffer ids = new StringBuffer(30);
     * for (Elokuva elokuva:elokuvat) // Kokeillaan for-silmukan toimintaa
     *   ids.append(" "+elokuva.getElokuvaID());
     * String tulos = " " + terminator1.getElokuvaID() + " " + terminator2.getElokuvaID() + " " + terminator1.getElokuvaID();
     * ids.toString() === tulos;
     * ids = new StringBuffer(30);
     * for (Iterator<Elokuva>  i=elokuvat.iterator(); i.hasNext(); ) { // ja iteraattorin toimintaa
     *   Elokuva elokuva = i.next();
     *   ids.append(" "+elokuva.getElokuvaID());
     * }
     * ids.toString() === tulos;
     * Iterator<Elokuva>  i=elokuvat.iterator();
     * i.next() == terminator1  === true;
     * i.next() == terminator2  === true;
     * i.next() == terminator1  === true;
     * i.next();  #THROWS NoSuchElementException
     * </pre>
     */
    public class ElokuvatIterator implements Iterator<Elokuva> {
        private int kohdalla = 0;


        /**
         * Onko olemassa vielä seuraavaa elokuvaa
         * @see java.util.Iterator#hasNext()
         * @return true jos on vielä elokuvia
         */
        @Override
        public boolean hasNext() {
            return kohdalla < getLkm();
        }


        /**
         * Annetaan seuraava elokuva
         * @return seuraava elokuva
         * @throws NoSuchElementException jos seuraava alkiota ei enää ole
         * @see java.util.Iterator#next()
         */
        @Override
        public Elokuva next() throws NoSuchElementException {
            if ( !hasNext() ) throw new NoSuchElementException("Ei ole");
            return anna(kohdalla++);
        }


        /**
         * Tuhoamista ei ole toteutettu
         * @throws UnsupportedOperationException aina
         * @see java.util.Iterator#remove()
         */
        @Override
        public void remove() throws UnsupportedOperationException {
            throw new UnsupportedOperationException("Ei voi poisteta");
        }
    }


    /**
     * Palautetaan iteraattori elokuvista
     * @return elokuva iteraattori
     */
    @Override
    public Iterator<Elokuva> iterator() {
        return new ElokuvatIterator();
    }


    /**
     * Palauttaa "taulukossa" hakuehtoon vastaavien elokuvien viitteet
     * @param hakuehto hakuehto
     * @param k milla kentalla jarjesetaan
     * @return tietorakenne loytyneista elokuvista
     * @example <pre name="test">
     * #THROWS SailoException
     *  Elokuvat elokuvat = new Elokuvat();
     *  Elokuva e1 = new Elokuva(); e1.parse("1|Terminator|James Cameron|1984|");
     *  Elokuva e2 = new Elokuva(); e2.parse("2|Terminator 2|James Cameron|1992|");
     *  Elokuva e3 = new Elokuva(); e3.parse("3|Terminator 3|James Cameron|2003|");
     *  Elokuva e4 = new Elokuva(); e4.parse("4|Pulp Fiction|Quentin Tarantino|1994|");
     *  elokuvat.lisaa(e1); elokuvat.lisaa(e2); elokuvat.lisaa(e3); elokuvat.lisaa(e4);
     *  List<Elokuva> loytyneet;
     *  loytyneet = (List<Elokuva>)elokuvat.etsi("Termi", 1);
     *  loytyneet.size() === 3;
     *  loytyneet = (List<Elokuva>)elokuvat.etsi("pulp", 1);
     *  loytyneet.size() === 1;
     * </pre>
     */
    public Collection<Elokuva> etsi(String hakuehto, int k) {
        List<Elokuva> loytyneet = new ArrayList<Elokuva>();

        // Muutetaan ehto pieneksi
        String ehto = hakuehto.toLowerCase();
        String elokuvanNimi;

        // Tarkistetaan kentta
        int hk = k;
        if ( hk < 0 ) hk = 1;

        for (Elokuva elokuva : this) {
            elokuvanNimi = elokuva.getNimi().toLowerCase();
            if (elokuvanNimi.contains(ehto)) { // Katsotaan sisaltaako elokuva haettua nimea
                loytyneet.add(elokuva);
            }
        }
        Collections.sort(loytyneet, new Elokuva.Vertailija(hk));
        return loytyneet;
    }


    /**
     * Testiohjelma elokuville
     */
    public static void main() {
        Elokuvat elokuvat = new Elokuvat();

        try {
            elokuvat.lueTiedostosta("omakirjasto");

            Elokuva Terminator = new Elokuva();
            Terminator.rekisteroi();
            Terminator.taytaElokuva();
            elokuvat.lisaa(Terminator);

            elokuvat.tallenna("omakirjasto");
        } catch (SailoException ex) {
            System.out.println(ex.getMessage());
        }

        System.out.println("============= Elokuvat testi =================");
        // Annetaan elouvan numero jarjestyksessa
        for (int i = 0; i < elokuvat.getLkm(); i++) {
            Elokuva elokuva = elokuvat.anna(i); // Asetetaan i:des viite elokuvaan
            System.out.println("Elokuva nro: " + i);
            elokuva.tulosta(System.out);
        }

    }
}