package kirjasto;

import java.io.*;
import java.util.*;
/**
 * Genret, lisaa uuden genren
 * HUOM: Tällä hetkellä käyttöliittymässä ei ole uusien genrejen lisäys nappia,
 * eli ladataan valmiit genret tiedostosta
 * TODO: Genren lisäys napin voisi toteuttaa tulevaisuudessa helposti, sillä rekisteröinti on jo olemassa genreillä
 * @author Erno Pajala
 * @version 08.03.2025
 **/
public class Genret implements Iterable<Genre> {
    private boolean muutettu = false;

    /** Taulukko genreista */
    private final Collection<Genre> alkiot = new ArrayList<>();
    // Käytetaan yksinkertaista (collection) omisaisuutta arraylist rakenteesta

    /**
     * Genren alustaminen
     */
    public Genret() {
        // toistaiseksi ei tarvitse tehdä mitään
    }


    /**
     * Lisää uuden genren tietorakenteeseen.  Ottaa genren omistukseensa.
     * @param gen lisättävä genre. Huom tietorakenne muuttuu omistajaksi
     */
    public void lisaa(Genre gen) {
        alkiot.add(gen);
        muutettu = true;
    }


    /**
     * Lukee genret tiedostosta
     * @param hakemisto tiedoston hakemisto
     * @throws SailoException jos lukeminen epäonnistuu
     * @example
     * <pre name="test">
     * #THROWS SailoException
     * #import java.io.*;
     * // Testaa tallennus ja lukeminen
     * String testiHakemisto = "testidata";
     * Genret genret = new Genret();
     * Genre g1 = new Genre(), g2 = new Genre();
     * g1.rekisteroi(); g1.taytaGenre("Toiminta");
     * g2.rekisteroi(); g2.taytaGenre("Draama");
     * genret.lisaa(g1);
     * genret.lisaa(g2);
     * // Testaa kun tiedostoa ei voi lukea
     * genret.lueTiedostosta(testiHakemisto); #THROWS SailoException
     * genret.tallenna(testiHakemisto);
     * // Lue tiedot uusiin genreihin
     * Genret uudet = new Genret();
     * uudet.lueTiedostosta(testiHakemisto);
     * Iterator<Genre> i = uudet.iterator();
     * i.next().toString() === g1.toString();
     * i.next().toString() === g2.toString();
     * i.hasNext() === false;
     * // Siivoa testitiedosto
     * new File(testiHakemisto + "/genret.dat").delete();
     * new File(testiHakemisto).delete();
     * </pre>
     */
    public void lueTiedostosta(String hakemisto) throws SailoException {
        String tiedostonNimi = hakemisto + "/genret.dat";
        File ftiedosto = new File(tiedostonNimi);

        if (!ftiedosto.exists()) {
            throw new SailoException("Tiedostoa ei löydy: " + ftiedosto.getAbsolutePath());
        }

        try (Scanner scanner = new Scanner(new FileInputStream(ftiedosto))) {
            while ( scanner.hasNext() ) { // Jos seuraava olemassa jatka lukemista
                String s = scanner.nextLine();
                if (s ==  null || s.isEmpty()) continue; // Katsotaan lopetetaanko silmukka
                Genre gen = new Genre();
                gen.parse(s); // Otetaan rivilta tiedot ja laitetaan ne genrelle
                lisaa(gen); // Lisataan genre
            }
            muutettu = false;
        } catch ( FileNotFoundException e ) {
            throw new SailoException("Ei saa luettua tiedostoa " + tiedostonNimi);
        }
    }


    /**
     * Tallentaa genret tiedostoon
     * @param hakemisto Mihin hakemistoon tallennetaan
     * @throws SailoException jos talletus epäonnistuu
     */
    public void tallenna(String hakemisto) throws SailoException {
        if ( !muutettu ) return; // Ei tallenneta jos ei olla muutettu

        File kansio = new File(hakemisto);
        if (!kansio.exists()) {
            if (!kansio.mkdirs()) {
                throw new SailoException("Hakemiston luominen epäonnistui: " + kansio.getAbsolutePath());
            }
        }

        File ftiedosto = new File(hakemisto + "/genret.dat");
        try (PrintStream ps = new PrintStream(new FileOutputStream(ftiedosto)) ) {

            // Kaydaan lapi genret ja laitetaan ne rivi kerrallaan tiedostoon
            for (Genre gen : alkiot) {
                ps.println(gen.toString());
            }

        } catch (FileNotFoundException e) {
            throw new SailoException("Tiedosto " + ftiedosto.getAbsolutePath());
        }

        muutettu = false;
    }


    /**
     * Iteraattori kaikkien genrejen läpikäymiseen
     * @return genreiteraattori
     *
     * @example
     * <pre name="test">
     * #PACKAGEIMPORT
     * #import java.util.*;
     *  Genret genr = new Genret();
     *  Genre gen21 = new Genre(); genr.lisaa(gen21);
     *  Genre gen11 = new Genre(); genr.lisaa(gen11);
     *  Genre gen22 = new Genre(); genr.lisaa(gen22);
     *  Genre gen12 = new Genre(); genr.lisaa(gen12);
     *  Genre gen23 = new Genre(); genr.lisaa(gen23);
     *  Iterator<Genre> i2=genr.iterator();
     *  i2.next() === gen21;
     *  i2.next() === gen11;
     *  i2.next() === gen22;
     *  i2.next() === gen12;
     *  i2.next() === gen23;
     *  i2.next() === gen12;  #THROWS NoSuchElementException
     *  int n = 0;
     *  int genreIDs[] = {gen21.getGenreID(), gen11.getGenreID(), gen22.getGenreID(), gen12.getGenreID(), gen23.getGenreID()};
     *  for ( Genre gen : genr ) {
     *    gen.getGenreID() === genreIDs[n]; n++;
     *  }
     *  n === 5;
     *
     * </pre>
     */
    @Override
    public Iterator<Genre> iterator() {
        return alkiot.iterator();
    }


    /**
     * Haetaan kaikki elokuvan genret
     * @param genreIDs Elokuvan genreID lista, josta etsitään vastaavat genret
     * @return tietorakenne jossa viiteet löydetteyihin arvosteluihin
     * <pre name="test">
     * #import java.util.*;
     *  // Lisataan genreoliot
     *  Genret genree = new Genret();
     *  Genre gen21 = new Genre();
     *  gen21.rekisteroi();
     *  genree.lisaa(gen21);
     *  Genre gen11 = new Genre();
     *  gen11.rekisteroi();
     *  genree.lisaa(gen11);
     *  List<Genre> loytyneet; // Löytyneet genret
     *  List<Integer> genreIDs = new ArrayList<>();
     *  loytyneet = genree.annaGenret(genreIDs);
     *  loytyneet.size() === 0;
     *  // Lisataan etsittavat GenereID:t listaan
     *  genreIDs.add(1);
     *  genreIDs.add(2);
     *  // Etsitaan
     *  loytyneet = genree.annaGenret(genreIDs);
     *  loytyneet.size() === 2;
     *  loytyneet.get(0) == gen21 === true;
     *  loytyneet.get(1) == gen11 === true;
     *  loytyneet.get(1) == gen21 === false;
     * </pre>
     */
    public List<Genre> annaGenret(List<Integer> genreIDs) {
        List<Genre> loydetyt = new ArrayList<>(); // Löydetyt genret

        for (Genre gen : alkiot) { // Kaydaan kaikki genret läpi ja katsotaan löytyykö annettut elokuvan GenreIDt
            for (Integer genreID : genreIDs) { // käydään elokuvan genret läpi
                if (gen.getGenreID() == genreID) {
                    loydetyt.add(gen); // Jos genreID on sama kuin elokuvan genreID laitetaan genre löydettyihin
                }
            }
        }
        return loydetyt;
    }


    /**
     * Etsitaan kaikki kirjaston genret
     * @return palautetaan kaikki kirjaston genret
     */
    public List<Genre> annaKaikkiGenret() {
        List<Genre> allGenres = new ArrayList<>(); // Kaikki kirjaston genret

        // kaydaan lapi kaikki genre alkiot
        for (Genre gen : alkiot) {
            allGenres.add(gen);
        }
        return allGenres;
    }


    /**
     * Testiohjelma arvosteluille
     * @param args ei käytössä
     */
    public static void main(String[] args) {
        Genret genreja = new Genret();

        try {
            genreja.lueTiedostosta("omakirjasto");
            // Elokuvan GenreIDt tesiin, eli testissa elokuvalla on 2 genrea, joiden ID:t ovat [2, 5]
            List<Integer> movieGenreIDs = new ArrayList<>();
            movieGenreIDs.add(2);
            movieGenreIDs.add(5);

            System.out.println("============= Genre testi =================");
            List<Genre> genret2 = genreja.annaGenret(movieGenreIDs);
            for (Genre gen : genret2) {
                gen.tulosta(System.out);
            }
        } catch (SailoException e) {
            throw new RuntimeException(e);
        }
    }
}
