package kirjasto;

import java.io.File;
import java.util.Collection;
import java.util.List;

/**
 * Kirjasto-luokka, joka huolehtii elokuvista
 *
 * @author Erno Pajala
 * @version 27.02.2025
 */
public class Kirjasto {
    private Elokuvat elokuvat = new Elokuvat();
    private Arvostelut arvostelut = new Arvostelut();
    private Genret genret = new Genret();


    /**
     * Lisataan uusi elokuva
     * @param elokuva lisattava elokuva
     * @throws SailoException jos lisaaminen ei onnistu
     */
    public void lisaa(Elokuva elokuva) throws SailoException {
        elokuvat.lisaa(elokuva);
    }


    /**
     * Lisätään uusi arvostelu
     * @param arvostelu lisattava arvostelu
     */
    public void lisaa(Arvostelu arvostelu) {
        arvostelut.lisaa(arvostelu);
    }


    /**
     * Lisätään uusi genre
     * @param genre lisattava genre
     */
    public void lisaa(Genre genre) {
        genret.lisaa(genre);
    }


    /**
     * Korvataan tai lisataan elokuva
     * @param elokuvaklooni milla korvataan
     * @throws SailoException jos tulee virheita
     */
    public void korvaaTaiLisaa(Elokuva elokuvaklooni) throws SailoException {
        elokuvat.korvaaTaiLisaa(elokuvaklooni);
    }


    /**
     * Kortvaa tai lissataan arvostelu
     * @param arvosteluklooni milla korvataan
     * @throws SailoException jos tulee virhe
     */
    public void korvaaTaiLisaa(Arvostelu arvosteluklooni) throws SailoException {
        arvostelut.korvaaTaiLisaa(arvosteluklooni);
    }


    /**
     * Annetaan elokuvan genret, elokuvan genreID:tten avulla
     * @param elokuva elokuva josta haetaan genreID:t
     * @return Löydetyt genret
     * <pre name="test">
     * #import java.util.*;
     *  Kirjasto kirjasto = new Kirjasto();
     *  Elokuva Terminator1 = new Elokuva();
     *  Elokuva Terminator2 = new Elokuva();
     *  Terminator1.rekisteroi(); Terminator2.rekisteroi();
     *  Terminator1.taytaElokuva();
     *  Genre gen1 = new Genre();
     *  Genre gen2 = new Genre();
     *  gen1.rekisteroi(); gen2.rekisteroi();
     *  gen1.taytaGenre("Sci-Fi"); gen2.taytaGenre("Toiminta");
     *  kirjasto.lisaa(gen1);
     *  kirjasto.lisaa(gen2);
     *  List<Genre> loytyneet;
     *  loytyneet = kirjasto.annaGenret(Terminator2);
     *  loytyneet.size() === 0;
     *  loytyneet = kirjasto.annaGenret(Terminator1);
     *  loytyneet.size() === 2;
     *  loytyneet.get(0) == gen1 === true;
     *  loytyneet.get(1) == gen2 === true;
     * </pre>
     */
    public List<Genre> annaGenret(Elokuva elokuva) {
        return genret.annaGenret(elokuva.getGenreIDs());
    }


    /**
     * Annetaan kaikki kirjaston genret
     * @return kaikki genret
     */
    public List<Genre> annaKaikkiGenret(){
        return genret.annaKaikkiGenret();
    }


    /**
     * Annetaan arvostelu elokuvan kohdalla ElokuvaID:n avulla
     * @param i elokuvan ID
     * @return Arvostelu kohdalla tai null
     * @throws IndexOutOfBoundsException jos i väärin
     */
    public Arvostelu annaArvostelu(int i) {
        return arvostelut.annaArvostelu(i);
    }


    /**
     * @return elokuvien maara
     */
    public int getElokuvia() throws IndexOutOfBoundsException{
        return elokuvat.getLkm();
    }


    /**
     * Antaa kirjaston i:n elokuvan
     * @param i monesko elokuva
     * @return elokuva paikassa i
     * @throws IndexOutOfBoundsException jos i väärin
     */
    public Elokuva annaElokuva(int i) throws IndexOutOfBoundsException {
        return elokuvat.anna(i);
    }


    /**
     * Palauttaa "Taulukossa" hakuehtoa vastaavien elokuvien viitteet
     * @param ehto hakuehto
     * @param k milla kentalla jarjesetaan
     * @return tietorakenne loytyneista elokuvista
     */
    public Collection<Elokuva> etsi(String ehto, int k) {
        return elokuvat.etsi(ehto, k);
    }


    /**
     * Lukee kirjaston tiedot tiedostosta
     * @param hakemisto kirjaston hakemiston nimi
     * @throws SailoException jos lukeminen epaonnistuu
     */
    public void lueTiedostosta(String hakemisto) throws SailoException {
        // Katsotaan ja luodaan uusi hakemisto jos ei ole olemassa
        File dir = new File(hakemisto);

        // Jos hakemisto on jo olemassa, korvataan luodut oletus genret tiedostossa olevilla genreilla
        if (dir.exists()) {
            genret = new Genret();
        }

        dir.mkdir();

        elokuvat = new Elokuvat(); // Jos luetaan olemassa oleva, niin tyhjennetaan se nain
        arvostelut = new Arvostelut();

        elokuvat.lueTiedostosta(hakemisto);
        arvostelut.lueTiedostosta(hakemisto);
        genret.lueTiedostosta(hakemisto);

    }


    /**
     * Tallettaa kirjaston tiedot tiedostoon
     * @param hakemisto kirjaston hakemiston nimi
     * @throws SailoException jos tallentaminen ei onnistu
     */
    public void tallenna(String hakemisto) throws SailoException {
        String virhe = "";
        try {
            elokuvat.tallenna(hakemisto);
        } catch (SailoException e) {
            virhe = e.getMessage();
        }

        try {
            arvostelut.tallenna(hakemisto);
        } catch (SailoException e) {
            virhe = e.getMessage();
        }

        try {
            genret.tallenna(hakemisto);
        } catch (SailoException e) {
            virhe = e.getMessage();
        }

        if ( !"".equals(virhe) ) { // Jos saadaan poikkeus jommasta kummasta, laitetaan se virheeseen ja tulostetaan se
            throw new SailoException(virhe);
        }
    }


    /**
     * Poistaa kirjastosta elokuvan ja elokuvan arvostelun
     * @param elokuva elokuva joka poistetaan
     * @return elokuva, joka poistettiin
     * <pre name="test">
     * #THROWS Exception
     *   Elokuvat elokuvat = new Elokuvat();
     *   Elokuva e1 = new Elokuva();
     *   Elokuva e2 = new Elokuva();
     *   e1.rekisteroi(); e2.rekisteroi();
     *   elokuvat.lisaa(e1); elokuvat.lisaa(e2);
     *   elokuvat.getLkm() === 2;
     *   int n = elokuvat.poista(e1.getElokuvaID());
     *   n === 1;
     *   elokuvat.getLkm() === 1;
     *   n = elokuvat.poista(e2.getElokuvaID());
     *   elokuvat.getLkm() === 0;
     *   n === 1;
     *   n = elokuvat.poista(3);
     *   n === 0;
     * </pre>
     */
    public int poista(Elokuva elokuva) {
        if (elokuva == null) return 0;
        int n = elokuvat.poista(elokuva.getElokuvaID());
        // Poistetaan elokuvan arvostelu
        arvostelut.poista(elokuva.getElokuvaID());
        return n;
    }


    /**
     * Poistaa kirjastosta arvostelu
     * @param arvostelu joka poistetaan
     * <pre name="test">
     * #THROWS Exception
     *   Arvostelut arvostelut = new Arvostelut();
     *   Arvostelu a1 = new Arvostelu(); Arvostelu a2 = new Arvostelu();
     *   a1.rekisteroi(1); a2.rekisteroi(2);
     *   arvostelut.lisaa(a1); arvostelut.lisaa(a2);
     *   arvostelut.annaArvostelu(1) === a1;
     *   arvostelut.annaArvostelu(2) === a2;
     *   arvostelut.poista(a2.getElokuvaID());
     *   arvostelut.annaArvostelu(1) === a1;
     *   arvostelut.annaArvostelu(2) === null;
     */
    public void poista(Arvostelu arvostelu) {
        if (arvostelu == null) return;
        arvostelut.poista(arvostelu.getElokuvaID());
    }


    /**
     * Testiohjelma kirjastolle
     * @param args ei käytössä
     */
    public static void main(String[] args) {
        Kirjasto kirjasto = new Kirjasto();

        // kirjasto.lueTiedostosta("movies");
        Elokuva Terminator = new Elokuva(), Terminator2 = new Elokuva();
        Terminator.rekisteroi();
        Terminator.taytaElokuva();
        Terminator2.rekisteroi();
        Terminator2.taytaElokuva();

        // kirjasto.lueTiedostosta("genres")
        Genre gen1 = new Genre(), gen2 = new Genre();
        gen1.rekisteroi();
        gen2.rekisteroi();
        gen1.taytaGenre("Toimita");
        gen2.taytaGenre("Sci-fi");

        try {
            kirjasto.lisaa(Terminator); // tarkisteteaan tarttuiko lisatyt elokuvat kirjastoon
            kirjasto.lisaa(Terminator2); // Tarkistetaan onko eri elokuvaa, ettei sama elokuva osoita samaan paikkaan
            kirjasto.lisaa(gen1);
            kirjasto.lisaa(gen2);
        } catch (SailoException e) {
            System.err.println(e.getMessage());
        }


        for (int i = 0; i < kirjasto.getElokuvia(); i++) {
            Elokuva elokuva = kirjasto.annaElokuva(i);

            List<Genre> genreet = kirjasto.annaGenret(elokuva);

            // Kaydaan kaikki elokuvien genret läpi
            for (Genre genre : genreet) {
                System.out.println("Genre ID: " + genre.getGenreID() + " Genre: " + genre.getGenreName());
            }

            elokuva.tulosta(System.out);
        }
    }
}